/* global utils, CommunicationManager, application, DeviceManager, defined */

(function() {


	"use strict";
	/**
	 * @classdesc Log facility for the application.
	 * @type Logger
	 * @version 1.8.1
	 * @since 1.8.1
	 */
	//jshint -W057
	window.Logger = new function _Logger() {

		/* +------------------------+ Private Variables +------------------------+ */
		
		/**
		 * Instance of IndexDB. 
		 * @version 1.8.1
		 * @since 1.8.1
		 */
		var v_db;
		
		/**
		 * Name of Logs db object store.
		 * @type {String} 
		 * @version 1.8.1
		 * @since 1.8.1
		 */
		var v_nameObjStoreLogs = "logs";

		/**
		 * Name of Statistics db object store.
		 * @type {String} 
		 * @version 1.8.1
		 * @since 1.8.1
		 */
		var v_nameObjStoreStats = "stats";
		
		var v_cmsLogCategory = Object.create(null);
	    Object.defineProperty(v_cmsLogCategory, 'AUDIT', {
		  enumerable: false,
		  configurable: false,
		  writable: false,
		  value: 'audit'
		});

	    Object.defineProperty(v_cmsLogCategory, 'ERROR', {
		  enumerable: false,
		  configurable: false,
		  writable: false,
		  value: 'error'
		});
		/* +------------------------+ Private Function  +------------------------+ */

		/**
	     * Add log entry to CMS log.
	     * @param {Object} entry An object with the folowing fields.
	     * <table>
	     * <tr><th>Type</th><th>Field</th><th>Description</th><th>Optional</th></tr> 
	     * <tr><td>"error" | "audit"</td><td>category</td><td>Either error or audit.</td><td>No</td></tr> 
	     * <tr><td></td><td></td><td>Audit messages are discarded unless auditing<br> is switched on globally and on the display.</td><td></td></tr> 
	     * <tr><td>String</td><td>message</td><td>The log message.</td><td>No</td>
	     * <tr><td>String</td><td>scheduleID</td><td>The assosiated schedule id.</td><td>Yes</td></tr> 
	     * <tr><td>String</td><td>layoutID</td><td>The assosiated layout id.</td><td>Yes</td></tr> 
	     * <tr><td>String</td><td>mediaID</td><td>The assosiated mediaID id.</td><td>Yes</td></tr> 
	     * <tr><td>String</td><td>type</td><td>The type.</td><td>Yes</td></tr> 
	     * <tr><td>String</td><td>method</td><td>The method.</td><td>Yes</td></tr> 
	     * <tr><td>String</td><td>thread</td><td>The Thread that the log message executed on.</td><td>Yes</td></tr> 
	     * <tr><td>Boolean</td><td>onlyOnce</td><td>Don't put the same message more than once if already exists in the log to send.</td><td>Yes</td></tr> 
	     * </table>
	     * @returns None
	     * @example cmsLog({category:"error",message:"Error message"});
	     */
	    
	    /**
	     * Create a new log entry.
	     * @param {[Logger.cmsLog.CATEGORY.ERROR | Logger.cmsLog.CATEGORY.AUDIT]} category Either error or audit. Audit messages are discarded unless auditing<br> is switched on globally and on the display.
	     * @param {String} message The log message.
	     * @returns {Logger.logEntry} Return a new log entry.
	     */
	    function logEntry(category, message){

            if (category !== v_cmsLogCategory.ERROR && category !== v_cmsLogCategory.AUDIT){
                throw new TypeError("Category must be 'error' or 'audit'");
            }
	    	
            if (!defined(message)){
                throw new TypeError("Message must be set!");
            }
	        this.date = utils.getISODate();
	        this.category = category;
	        this.message = message;
	        this.onlyOnce = false;
	        
	        /**
	         * Get the DOM representation of the log entry.
	         * @returns {Object} Return the DOM equivelant to the log entry.
	         */
	        this.constructor.prototype.toElement = function(){
	            var doc = document.implementation.createDocument(null, "l", null);

	            var de = doc.createElement("log");
	            de.setAttribute("date",this.date);
	            de.setAttribute("category",this.category);

	            var me = doc.createElement("message");
	            me.textContent = this.message;
	            de.appendChild(me);

	            if (this.scheduleID !== undefined){
	                var si = doc.createElement("scheduleID");
	                si.textContent = this.scheduleID;
	                de.appendChild(si);
	            }

	            if (this.layoutID !== undefined){
	                var li = doc.createElement("layoutID");
	                li.textContent = this.layoutID;
	                de.appendChild(li);
	            }

	            if (this.mediaID !== undefined){
	                var mi = doc.createElement("mediaID");
	                mi.textContent = this.mediaID;
	                de.appendChild(mi);
	            }

	            if (this.type !== undefined){
	                var ty = doc.createElement("type");
	                ty.textContent = this.type;
	                de.appendChild(ty);
	            }

	            if (this.method !== undefined){
	                var met = doc.createElement("method");
	                met.textContent = this.method;
	                de.appendChild(met);
	            }

	            if (this.thread !== undefined){
	                var th = doc.createElement("thread");
	                th.textContent = this.thread;
	                de.appendChild(th);
	            }
	            
	            return de;
	        };
	        
	        /**
	         * Compare if two entry logs are equal.
	         * @param {type} compareObj The entry log to compare to.
	         * @returns {Boolean} True if equal else false.
	         */
	        this.constructor.prototype.equals = function(compareObj){
	            var cmpAttribs = ["category", "message", "method", "thread", "type", "scheduleID", "layoutID", "mediaID"];
	            
	            for (var i = 0, max = cmpAttribs.length; i < max; i++) {
	                if (this[cmpAttribs[i]] !== compareObj[cmpAttribs[i]]){
	                    return false;
	                }
	            }
	            return true;
	        };
	        
	        /**
	         * Search into indexDB if this log entry exists.
	         * @param {function} successCb Callback function with result. True if exist else false.
	         * @param {function} failureCb Callback function in caes of an error.
	         * @returns {undefined} undefined.
	         */
	        this.constructor.prototype.exists = function(successCb, failureCb){
	        	var transaction = v_db.transaction([v_nameObjStoreLogs], "readonly");
	    		var objectStore = transaction.objectStore(v_nameObjStoreLogs);
	    		var request = objectStore.openCursor();
	        	var current = this;
	        	
	    		request.onerror = function(event){
	    			console.error(event);
	    			if (failureCb){
	    				failureCb(event);
	    			}
	    		};
	    		
	    		request.onsuccess = function(event) {
	    		    var cursor = event.target.result;
	    		    if(cursor) {
	    		    	var log = cursor.value;
	    		    	if (current.equals(log)){
	    		    		successCb(true);
	    		    	}else{
		  	              	cursor.continue();
	    		    	}
	    		    } else {
		    			if (successCb){
    	                    successCb(false);
    	                    return;
    	                }
	    		    }
	    		};
	        };
	    }

	    function prepareDB(){
	    	var req = window.indexedDB.open('xibo',1);
	    	
	    	req.onerror = function(ev)
	    	{
	    		console.error(ev);
    		};
    		
    		req.onsuccess = function(ev)
    		{
    			console.debug('Database initialized!');
    			v_db = ev.target.result;
    		};
    		
    		req.onupgradeneeded = function(ev)
    		{
    			//console.info(ev);
    			var db = ev.target.result;
    			
    			db.onerror = function(ug_ev) {
    			      console.error(ug_ev);
    			};
    			
    			if (ev.oldVersion === 0){
    				db.createObjectStore(v_nameObjStoreLogs, { autoIncrement : true });
    				db.createObjectStore(v_nameObjStoreStats, { autoIncrement : true });
    			}
    			
    		};
	    }
	    
	    prepareDB();
		/* +------------------------+ Public Functions  +------------------------+ */

		/**
	     * Add log entry to CMS log.
	     * @param {Object} entry An object with the folowing fields.
	     * <table>
	     * <tr><th>Type</th><th>Field</th><th>Description</th><th>Optional</th></tr> 
	     * <tr><td>"error" | "audit"</td><td>category</td><td>Either error or audit.</td><td>No</td></tr> 
	     * <tr><td></td><td></td><td>Audit messages are discarded unless auditing<br> is switched on globally and on the display.</td><td></td></tr> 
	     * <tr><td>String</td><td>message</td><td>The log message.</td><td>No</td>
	     * <tr><td>String</td><td>scheduleID</td><td>The assosiated schedule id.</td><td>Yes</td></tr> 
	     * <tr><td>String</td><td>layoutID</td><td>The assosiated layout id.</td><td>Yes</td></tr> 
	     * <tr><td>String</td><td>mediaID</td><td>The assosiated mediaID id.</td><td>Yes</td></tr> 
	     * <tr><td>String</td><td>type</td><td>The type.</td><td>Yes</td></tr> 
	     * <tr><td>String</td><td>method</td><td>The method.</td><td>Yes</td></tr> 
	     * <tr><td>String</td><td>thread</td><td>The Thread that the log message executed on.</td><td>Yes</td></tr> 
	     * <tr><td>Boolean</td><td>onlyOnce</td><td>Don't put the same message more than once if already exists in the log to send.</td><td>Yes</td></tr> 
	     * </table>
	     * @returns None
	     * @example cmsLog({category:"error",message:"Error message"});
	     */
	    this.cmsLog = function(entry){
	        
	        if (!defined(entry)){
	            console.error("CMS log entry not defined!");
	        }

	        if (!entry.hasOwnProperty("category")){
	            console.error("CMS log entry has no category defined!");
	        }
	        if (!entry.hasOwnProperty("message")){
	            console.error("CMS log entry has no message defined!");
	        }

	        var v_logEntry = new logEntry(entry.category, entry.message);
	        
	        if (entry.hasOwnProperty("scheduleID")){
	            v_logEntry.scheduleID = entry.scheduleID;
	        }
	        
	        if (entry.hasOwnProperty("layoutID")){
	            v_logEntry.layoutID = entry.layoutID;
	        }

	        if (entry.hasOwnProperty("mediaID")){
	            v_logEntry.mediaID = entry.mediaID;
	        }

	        if (entry.hasOwnProperty("type")){
	            v_logEntry.type = entry.type;
	        }

	        if (entry.hasOwnProperty("method")){
	            v_logEntry.method = entry.method;
	        }
	        
	        if (entry.hasOwnProperty("thread")){
	            v_logEntry.thread = entry.thread;
	        }

	        if (entry.hasOwnProperty("onlyOnce")){
	            v_logEntry.onlyOnce = entry.onlyOnce;
	        }
	        
	        if (v_logEntry.category === "audit"){
	            console.info("Audit", v_logEntry);
	        }else{
	            console.warn("Error", v_logEntry);
	        }
	        
	        
	        var addLogEntry = function(le){
	        	var transaction = v_db.transaction([v_nameObjStoreLogs], "readwrite");
		        
		        transaction.onerror = function() {
		        	console.error('Error adding log', transaction.error);
		        };
		        
		        var objectStore = transaction.objectStore(v_nameObjStoreLogs);
		        var objectStoreRequest = objectStore.add(le);
		        
		        objectStoreRequest.onsuccess = function() {
		        	console.debug('Log added', le);
		        };

		        objectStoreRequest.onerror = function(event) {
		        	console.debug('Log add failed', event);
		        };
		        
	        };
	        
	        var checkOnlyOnce = function(){
		        //Find if onlyOnce is set
		        if (v_logEntry.onlyOnce){
		        	v_logEntry.exists(
	        			function(result){
	        				if (!result){
	        					addLogEntry(v_logEntry);
	        				}
	        			},function(err){
	        				throw err;
	        			});
		                
		        }else{
		        	addLogEntry(v_logEntry);
		        }
	        };
	        
	        
	        var v_logLevel = DeviceManager.Config.getLogLevel();
	        
	        switch (v_logLevel) {
            case 'error':
            	if (v_logEntry.category === "error"){
            		checkOnlyOnce();
            	}
                break;
            case 'audit':
                checkOnlyOnce();
                break;
            default:
                //Do nothing
                break;
	        }
	        
	        application.statusPage.addLogMessage(v_logEntry);
	    };
	    
	    Object.defineProperty(this.cmsLog, 'CATEGORY', {
		  enumerable: false,
		  configurable: false,
		  writable: false,
		  value: v_cmsLogCategory
		});

	    /**
	     * Submit the log to the cms.
	     * @param {Function} successCb Function to call on success.
	     * @param {Function} failureCb Function to call on failure.
	     * @returns None
	     */
	    this.submitLog = function(successCb, failureCb){
	    	
	    	
	    	var emptyLogs = function(){
	    		var transaction = v_db.transaction([v_nameObjStoreLogs], "readwrite");
	    		var objectStore = transaction.objectStore(v_nameObjStoreLogs);
	    		var objectStoreRequest = objectStore.clear();
	    		objectStoreRequest.onsuccess = function(event) {
		    		if (successCb){
	                    successCb();
	                }
	    		};
	    		objectStoreRequest.onerror = function(event) {
	    			if (failureCb){
	    				failureCb(event);
	    			}
	    		};
	    	};
	    	
	    	var sendLogs = function(){
	    		var doc = document.implementation.createDocument(null, "logs", null);
	    		var haveLogs = false;
	    		
	    		var transaction = v_db.transaction([v_nameObjStoreLogs], "readonly");
	    		var objectStore = transaction.objectStore(v_nameObjStoreLogs);
	    		var request = objectStore.openCursor();
	    		request.onerror = function(event){
	    			console.error(event);
	    			if (failureCb){
	    				failureCb(event);
	    			}
	    		};
	    		
	    		request.onsuccess = function(event) {
	    		    var cursor = event.target.result;
	    		    if(cursor) {
	    		    	haveLogs = true;
	    		    	
	    		    	var logObj = Object.create(logEntry.prototype);
	    		    	var log = Object.assign(logObj, cursor.value);
	    		    	
	    		    	doc.firstChild.appendChild(log.toElement());
	  	              	
	  	              	cursor.continue();
	    		    } else {
	    		    	if (haveLogs){
	    		    		var l = (new XMLSerializer()).serializeToString(doc);
	    			        console.debug("Submit log:", l);
	    			        CommunicationManager.sendLog(emptyLogs,failureCb, l);
	    		    	}else{
    		    			if (successCb){
	    	                    successCb();
	    	                    return;
	    	                }
	    		    	}
	    		    }
	    		};
	    	};
	    	
	    	var transaction = v_db.transaction([v_nameObjStoreLogs], "readonly");
	    	var request = transaction.objectStore(v_nameObjStoreLogs).count();
	    	
	    	request.onsuccess = function(ev) {
	    		if (request.result > 0 ){
	    			sendLogs();
	    		}else{
	    	    	if (successCb){
	                    successCb();
	                    return;
	                }
	    		}
	    	};
	    	
	    	request.onerror = function(event){
    			console.error(event);
    			if (failureCb){
    				failureCb(event);
    			}
    		};
	    };

	    /**
	     * Add stat entry.
	     * @param {Object} stat An object with the folowing fields.
	     * <table>
	     * <tr><th>Type</th><th>Field</th><th>Description</th><th>Optional</th></tr> 
	     * <tr><td>String</td><td>fromdt</td><td>The ISO date that the layout/media started playing.</td><td>No</td></tr> 
	     * <tr><td>String</td><td>todt</td><td>The ISO date that the layout/media finished playing.</td><td>No</td>
	     * <tr><td>String</td><td>scheduleid</td><td>The ID of the schedule that caused the Layout/Media to be shown.</td><td>No</td></tr> 
	     * <tr><td>String</td><td>layoutid</td><td>The assosiated layout id.</td><td>Yes</td></tr> 
	     * <tr><td>String</td><td>mediaid</td><td>The assosiated mediaID id.</td><td>Yes</td></tr> 
	     * <tr><td colspan="4"><i>Either <b>layoutid</b> or <b>mediaid</b> must be set.</i></td></tr>
	     * </table>
	     * @returns None
	     * @see For ISO date see utils.getISODate()
	     * @example Logger.logStat({fromdt:utils.getISODate(),
	                        todt:utils.getISODate(),
	                        scheduleid:"5", 
	                        layoutid:"1"
	                       });<br>
	        Logger.logStat({fromdt:utils.getISODate(),
	                        todt:utils.getISODate(),
	                        scheduleid:"5", 
	                        mediaid:"1"
	                       });
	     */
	    this.logStat = function(stat){
	        
	        if (!DeviceManager.Config.getOptions().hasOwnProperty("statsEnabled"))
	        {
	            return;
	        }
	        
	        if (DeviceManager.Config.getOptions().statsEnabled == false)
	        {
	            return;
	        }
	        
	        if (!defined(stat)){
	            console.error("Statistic not defined!");
	        }

	        if (!stat.hasOwnProperty("layoutid")){
	            console.error("No layout id defined!");
	        }
	        
	        if (!stat.hasOwnProperty("fromdt")){
	            console.error("No from date defined!");
	        }

	        if (!stat.hasOwnProperty("todt")){
	            console.error("No to date defined!");
	        }

	        if (!stat.hasOwnProperty("scheduleid")){
	            console.error("No schedule id defined!");
	        }
	        
	        if (stat.hasOwnProperty("mediaid")){
	            stat.type = "media";
	        } else {
	            stat.type = "layout";
	        }
	        
	        var transaction = v_db.transaction([v_nameObjStoreStats], "readwrite");
	        
	        transaction.onerror = function(event) {
	        	console.error('Error adding stat', transaction.error);
	        };
	        
	        
	        var objectStore = transaction.objectStore(v_nameObjStoreStats);
	        var objectStoreRequest = objectStore.add(stat);
	        
	        objectStoreRequest.onsuccess = function(event) {
	        	console.debug('Stat added', stat);
	        };

	        objectStoreRequest.onerror = function(event) {
	        	console.debug('Stat add failed', event);
	        };
	        
	    };

	    /**
	     * Submit the log to the cms.
	     * @param {Function} successCb Function to call on success.
	     * @param {Function} failureCb Function to call on failure.
	     * @returns None
	     */
	    this.submitStats = function(successCb, failureCb){

	    	var emptyStats = function(){
	    		var transaction = v_db.transaction([v_nameObjStoreStats], "readwrite");
	    		var objectStore = transaction.objectStore(v_nameObjStoreStats);
	    		var objectStoreRequest = objectStore.clear();
	    		objectStoreRequest.onsuccess = function(event) {
		    		if (successCb){
	                    successCb();
	                }
	    		}
	    		objectStoreRequest.onerror = function(event) {
	    			if (failureCb){
	    				failureCb(event);
	    			}
	    		};
	    	};
	    	
	    	var sendStats = function(){
	    		var doc = document.implementation.createDocument(null, "stats", null);
	    		var haveStats = false;
	    		
	    		var transaction = v_db.transaction([v_nameObjStoreStats], "readonly");
	    		var objectStore = transaction.objectStore(v_nameObjStoreStats);
	    		var request = objectStore.openCursor();
	    		request.onerror = function(event){
	    			console.error(event);
	    			if (failureCb){
	    				failureCb(event);
	    			}
	    		};
	    		
	    		request.onsuccess = function(event) {
	    		    var cursor = event.target.result;
	    		    if(cursor) {
	    		    	haveStats = true;
	    		    	var stat = cursor.value;
	                    if (stat.hasOwnProperty('layoutid') && stat.layoutid!=='splash'){
		    		    	var de = doc.createElement("stat");
		    		    	for (var prop in stat) {
		    		    		de.setAttribute(prop,stat[prop]);
		    		    	}
			        	
		  	              	doc.firstChild.appendChild(de);
	                    }
	                    
	  	              	cursor.continue();
	    		    } else {
	    		    	if (haveStats){
	    		    		var l = (new XMLSerializer()).serializeToString(doc);
	    		    		console.debug("Submit stats:",l);
	    		    		CommunicationManager.sendStats(emptyStats,failureCb, l);
	    		    	}else{
    		    			if (successCb){
	    	                    successCb();
	    	                    return;
	    	                }
	    		    	}
	    		    }
	    		};
	    	};
	    	
	    	var transaction = v_db.transaction([v_nameObjStoreStats], "readonly");
	    	var request = transaction.objectStore(v_nameObjStoreStats).count();
	    	
	    	request.onsuccess = function(ev) {
	    		if (request.result > 0 ){
	    			sendStats();
	    		}else{
	    	    	if (successCb){
	                    successCb();
	                    return;
	                }
	    		}
	    	};
	    	
	    	request.onerror = function(event){
    			console.error(event);
    			if (failureCb){
    				failureCb(event);
    			}
    		};
	    };
	}();
}());
